(*****************************************************************************
 *
 * Type inference for a small subset of ML
 * Mainly written by Sorin Lerner, Fall 2000
 * Slightly edited by Craig Chambers, Fall 2000
 *
 * See the test_type_infer function at the end for the primary entry
 * point to invoking the type inferencer
 *
 *****************************************************************************)

open MLAst  (* The AST representation of ML programs, defined in ml_ast.ml *)


(*
 * This implemetation is incomplete. The parts that are missing will throw
 * the NotImplemented exception.
 *)

exception NotImplemented

(*****************************************************************************
 * Type declarations
 *****************************************************************************)

(*
 * This type is used for type variable identifiers, type constructor
 * identifiers, and user identifiers (ie: user variables)   
 *)

type Identifier = string


(* 
 * A type expression can either be:
 *
 *    1) a type variable        
 *    2) a type constructor with n type expressions as arguments.   
 *    3) a polymorphic type of the form ForAll V1..Vn: TypeExp
 *
 * Basic types such as int and bool are represented as a type
 * constructor with no arguments (eg: TypeCons("bool", nil)).
 *
 * Lists, functions, and tuples are all represented as type constructors.
 * The identifiers of the type constructors are "list", "->", and "*",
 * respectively.
 *)

datatype TypeExp = 
   TypeVar of Identifier |
   TypeCons of Identifier * TypeExp list |
   PolymorphicType of TypeExp list * TypeExp



(*  
 * Map from user identifiers to type expressions.
 *) 

type TypeMap = (Identifier * TypeExp) list


(*  
 * Map that represents a substitution. Each tuple is a replacement
 * rule: the first element in the tuple is a type variable, and the 
 * second is a type expression that should replace the type variable.     
 *) 

type SubstitutionMap = (TypeExp * TypeExp) list


(* 
 * A type environment is a pair containing:
 *     1. A type map from user variables to type expressions
 *     2. A substitution map that needs to be applied to type expressions
 *        in the type environment to get the final type expression.
 *)
    
type TypeEnv = TypeMap * SubstitutionMap

        
(* TypeError is raised if there is a type error in the user's program *)
exception TypeError


(*****************************************************************************
 * Definition of basic types (int, bool, string) and of the top level
 * type map  
 *****************************************************************************)

(*  
 * Get a new type variable. Uses side effects to generate unique variable ids.
 *)

val lastVarId = ref 0
fun newTypeVar() = 
( 
   lastVarId := !lastVarId + 1; 
   TypeVar("`var" ^ Int.toString(!lastVarId))
)

(*
 * Base types
 *)

val intType = TypeCons("int", nil);
val boolType = TypeCons("bool", nil);
val stringType = TypeCons("string", nil);


(*
 * Useful functions for creating function types, tuple types, and
 * polymorphic types.
 *)

fun createFunType(argType, resType) =
       TypeCons("->", [argType, resType])

fun createTupleType(x::nil) = x
|   createTupleType(typeList) = TypeCons("*", typeList)

fun createListType(listType) = 
       TypeCons("list", [listType]);


(*  
 *  isVarInList(list, var) -- returns whether or not the given type   
 *  variable is in the list of variables.
 *) 

fun isVarInList(nil, _) = false
|   isVarInList(TypeVar(varId)::xs, TypeVar(varToSearchFor)) =
        if varId = varToSearchFor then true 
        else isVarInList(xs, TypeVar(varToSearchFor));

fun createPolymorphicType(typeExp) =
   let
      fun createListOfFreeVars(typeVar as TypeVar(_), listOfFreeVars) =
             if isVarInList(listOfFreeVars, typeVar) then listOfFreeVars
             else typeVar::listOfFreeVars
      |   createListOfFreeVars(TypeCons(_, consList), 
                               listOfFreeVars) =
             foldl (fn (typeExp, listOfFreeVars) => 
                      createListOfFreeVars(typeExp, 
                                           listOfFreeVars))
                   listOfFreeVars
                   consList
      val listOfFreeVars = createListOfFreeVars(typeExp, nil)
   in
      if length(listOfFreeVars) = 0 then typeExp
      else PolymorphicType(listOfFreeVars, typeExp)
   end


(*
 * Converting types, maps, substitutions, and environments to their
 * string representation.
 *)

fun typeToString(TypeVar(varId)) = varId
  | typeToString(TypeCons("->", [dom,rng])) =
       "(" ^ typeToString(dom) ^ " -> " ^ typeToString(rng) ^ ")"
  | typeToString(TypeCons("*", consList)) =
       "(" ^ typeListToString(consList, " * ") ^ ")"
  | typeToString(TypeCons(consId, [])) = consId
  | typeToString(TypeCons(consId, [argType])) =
       typeToString(argType) ^ " " ^ consId
  | typeToString(TypeCons(consId, consList)) =
       "(" ^ typeListToString(consList, ", ") ^ ") " ^ consId
  | typeToString(PolymorphicType(quantifiedVars, typeExp)) = 
       "ForAll " ^ typeListToString(quantifiedVars, ",") ^ ": " ^ 
       typeToString(typeExp)

and typeListToString(typeList, sep) = 
      foldr (fn (elmt, foldResult) => 
	       let val es = typeToString(elmt)
	       in if foldResult = "" then es else es ^ sep ^ foldResult end)
             ""
             typeList


fun typeMapToString(typeMap) =
       "( " ^ 
        (foldr (fn ((userIdent, typeExp), foldResult) => 
                 "(" ^ userIdent ^ " ==> " ^ typeToString(typeExp) ^ 
                 ") " ^ foldResult)
             ""
             typeMap) ^
       " )"


fun substitutionMapToString(substitutionMap) =
       "( " ^ 
       (foldr (fn ((TypeVar(varId), typeExp), foldResult) => 
                 "(" ^ varId ^ " ==> " ^ typeToString(typeExp) ^ 
                 ") " ^ foldResult)
             ""
             substitutionMap) ^
       " )"


fun envToString((typeMap, substitutionMap)) =
       "typeMap: " ^ typeMapToString(typeMap) ^ "\n" ^
       "substitutionMap: " ^ substitutionMapToString(substitutionMap) ^ "\n"


(*****************************************************************************
 * Top level type map.  Defines the types of the predefined functions.
 *****************************************************************************)

val topLevelTypeMap =  [
   ("+", createFunType(createTupleType([intType, intType]), intType)),
   ("-", createFunType(createTupleType([intType, intType]), intType)),
(* add more cases here if additional primops defined *)
   ("cons", let val typeVar = newTypeVar()
                val listType = createListType(typeVar)
            in
               createPolymorphicType(
                 createFunType(createTupleType([typeVar, listType]), listType))
            end),
   ("null", createPolymorphicType(
                  createFunType(createListType(newTypeVar()), boolType))),
   ("hd", let val typeVar = newTypeVar()
          in
             createPolymorphicType(
                createFunType(createListType(typeVar), typeVar))
          end),
   ("tl", let val listType = createListType(newTypeVar())
          in
             createPolymorphicType(createFunType(listType, listType))
          end),
   ("nil", createPolymorphicType(createListType(newTypeVar())))
];


fun createTopLevelEnv() = (topLevelTypeMap, nil);


(*****************************************************************************
 * Generally useful functions. Only one for now.
 *****************************************************************************)

(*
 * mapAndFoldl: does a foldl and a map at the same time, returning a tuple
 * which containts the new list and the result of the fold.
 *)

fun mapAndFoldl(f, b, list) =
       let
          fun mapAndFoldlHelper(f, nil, foldResult, mapResult) =
                 (rev(mapResult), foldResult)
          |   mapAndFoldlHelper(f, x::xs, foldResult, mapResult) =
                 let
                    val (newElmt, newFoldResult) = f(x, foldResult)
                 in
                    mapAndFoldlHelper(f, xs, 
                                      newFoldResult, 
                                      newElmt::mapResult)
                 end
       in
          mapAndFoldlHelper(f, list, b, nil)
       end


(*****************************************************************************
 * Operations to instantiate a type expression (which involves making
 * fresh copies of the type variables for polymorphic types.
 *****************************************************************************)

fun getAlreadyCreatedVar(_, nil) = NONE
|   getAlreadyCreatedVar(typeVar as TypeVar(_), (old, new)::xs) =
       if (typeVar = old) then SOME new
       else getAlreadyCreatedVar(typeVar, xs);
          
fun getCopyOfTypeVar(typeVar, varsAlreadyCreated) = 
       (case (getAlreadyCreatedVar(typeVar, varsAlreadyCreated)) of
          SOME new => (new, varsAlreadyCreated) |
          NONE => let val new = newTypeVar()
                  in
                     (new, (typeVar, new)::varsAlreadyCreated)
                  end)


fun instTypeHelper(typeVar as TypeVar(varId), 
                   quantifiedVars, 
                   varsAlreadyCreated) =
      if not(isVarInList(quantifiedVars, typeVar)) then 
         (typeVar, varsAlreadyCreated)
      else
         getCopyOfTypeVar(typeVar, varsAlreadyCreated)
|   instTypeHelper(TypeCons(consId, consList),
                   quantifiedVars,
                   varsAlreadyCreated) =
       let
          val (newList, newVarsAlreadyCreated) = 
             mapAndFoldl(fn (elmt, reduceResult) => 
                            instTypeHelper(elmt, 
                                           quantifiedVars, 
                                           reduceResult),
                          varsAlreadyCreated, 
                          consList)
       in
          (TypeCons(consId, newList), newVarsAlreadyCreated)
       end
|   instTypeHelper(PolymorphicType(varList, typeExp),
                   quantifiedVars,
                   varsAlreadyCreated) =
       instTypeHelper(typeExp, 
                      List.concat([quantifiedVars, varList]), 
                      varsAlreadyCreated)

(*
 * Instantiate the given type expression. If the type expression is
 * polymorphic, this includes making copies of the variables 
 * that are bound to quantifiers in the expression.
 *)

fun instType(typeExp) =
   let val (result, _) = instTypeHelper(typeExp, nil, nil)
   in
      result
   end


(*****************************************************************************
 * Operations on the environment
 *****************************************************************************)

(*
 * getTypeFromMap(typeMap, typeEnv), retrieve the type of a user identifier    
 *)

fun getTypeFromMap(nil, _) = raise TypeError
|   getTypeFromMap((id,typeExp)::xs, idToSearchFor) =
       if id = idToSearchFor then typeExp
       else getTypeFromMap(xs, idToSearchFor);

(*
 * getType(typeEnv): typeExp -- retrieves the type of a user identifier from
 * the environment.
 *)

fun getType((typeMap, _): TypeEnv, 
            idToSearchFor: Identifier) = 
       getTypeFromMap(typeMap, idToSearchFor)


(*
 * Get the type, and instantiate it.
 *)

fun getAndInstType(typeEnv, idToSearchFor) = 
       instType(getType(typeEnv, idToSearchFor));


(*****************************************************************************
 * Substitution map operations
 *****************************************************************************)

(*
 * setSubsitution(substitutionMap, typeVarToAdd, newTypeExp): newSubstMap
 *
 * Adds a substitution rule to the given substitution, and returns the new
 * substitution. If a rule already exists for the given type variable, 
 * the rule is changed.
 *)

fun setSubstitution(substitutionMap, typeVarToAdd, newTypeExp) =
       let
          fun setSubstHelper(nil, _, _, true) = nil

          |   setSubstHelper(nil, varSearchedFor, newTypeExp, false) =
                 (varSearchedFor, newTypeExp)::nil

          |   setSubstHelper((TypeVar(varId), typeExp)::xs,
                              TypeVar(varSearchedFor), 
                              newTypeExp,
                              alreadyFound) =
                 let
                    val foundHere = (varId = varSearchedFor)
                    val rest = setSubstHelper(xs, 
                                              TypeVar(varSearchedFor), 
                                              newTypeExp,
                                              foundHere orelse alreadyFound)
                 in
                    if foundHere 
                       then (TypeVar(varSearchedFor), newTypeExp)::rest
                       else (TypeVar(varId), typeExp)::rest
                 end
        in
           setSubstHelper(substitutionMap, typeVarToAdd, newTypeExp, false)
        end
                       
(*
 * getSubstitution(substitutionMap, typeVar): typeExp OPTION
 *)

fun getSubstitution(nil, _) = NONE
|   getSubstitution((TypeVar(varId), typeExp)::xs, TypeVar(varSeachedFor)) = 
       if varId = varSeachedFor then SOME typeExp
       else getSubstitution(xs, TypeVar(varSeachedFor));

(*
 * applySubst(typeExp, substitutionMap) : (TypeExp, newSubstitutionMap)
 *
 * Applies the given substitution to typeExp repeatedly until it cannot
 * be applied anymore. The return value is the expanded type expression,
 * as well as a new pruned substitution map (eg: if 'a goes to 'b, 
 * which goes to 'c, then in the new map 'a goes to 'c)
 *)

fun
    applySubst(typeVar as TypeVar(varId), substitutionMap) = 
       (case getSubstitution(substitutionMap, typeVar) of
          NONE => (typeVar, substitutionMap) |
          SOME substitution => 
             let
                val (newType, newMap) = 
                   applySubst(substitution, substitutionMap)
             in
                (newType, setSubstitution(newMap, typeVar, newType)) 
             end)

|   applySubst(TypeCons(consId, consList), substitutionMap) =
       let
          val (newList, newSubstitutionMap) = mapAndFoldl(applySubst, 
                                                          substitutionMap,
                                                          consList)
       in
          (TypeCons(consId, newList), newSubstitutionMap)
       end

    
(*****************************************************************************
 * Check for a type variable in a type expression
 *****************************************************************************)

fun

    varOccursInTypeList(_, nil, _) = false

|   varOccursInTypeList(typeVar, x::xs, substMap) = 
       varOccursInTypeExp(typeVar, x, substMap) 
          orelse varOccursInTypeList(typeVar, xs, substMap)

and

    varOccursInTypeExp(typeToSearch as TypeVar(searchId), 
                       typeToLookIn,
                       substMap) =
       let val (expandedType, _) =
              applySubst(typeToLookIn, substMap)
       in
          case expandedType of
             TypeVar(prunedId) => searchId = prunedId |
             TypeCons(_, list) =>
                 varOccursInTypeList(typeToSearch, list, substMap)
       end



(*****************************************************************************
 * Unification
 *****************************************************************************)

(*  
 * unifyTypes(typeExp1, typeExp2, subtitutionMap): substitutionMap
 *  
 * Unifies the two given type expressions, and returns a new
 * substitution map.    
 *)

fun unifyTypes(typeExp1, typeExp2, substitutionMap): SubstitutionMap = 
       raise NotImplemented


(*****************************************************************************
 * Analyse an expression. Returns a type, and an environment.
 *****************************************************************************)

(*  
 * analyzeExp(expr, typeEnv): (typeExp, newTypeEnv)
 *      
 * Analyzes the given expression from the input language, and returns   
 * its type, as well as the new type environment. Note that the
 * new type environment contains a subsitution that needs to be
 * applied to the returned type expression in order to get the final
 * type.
 *)

(*
 * UnhandledLanguageConstruct is the exception raised if the user's
 * program uses a construct in ML that we don't handle type-inference for
 *)
exception UnhandledLanguageConstruct

(*
 * Identifier reference
 *) 
fun analyzeExp(var(varId), typeEnv): TypeExp * TypeEnv = raise NotImplemented

(*  
 * if-then-else
 *)
|   analyzeExp(if_else(test, iftrue, iffalse), typeEnv) = raise NotImplemented

(*
 * A function application
 *) 
|   analyzeExp(call(funExp, argExp), typeEnv) = raise NotImplemented

(*  
 * Handling a tuple expresion
 *) 
|   analyzeExp(tuple(expressions), typeEnv) = raise NotImplemented

(*  
 * Primitive operators are things like plus and times. You can find a list
 * of them in ml_ast.ml. You should treat them as function calls.
 *)
|   analyzeExp(primop(primOp, exp), typeEnv) = raise NotImplemented

(*  
 * Handling constants
 *)
|   analyzeExp(const(int(_)), typeEnv) = raise NotImplemented
|   analyzeExp(const(bool(_)), typeEnv) = raise NotImplemented
|   analyzeExp(const(string(_)), typeEnv) = raise NotImplemented

(*
 * Handled for a bonus
 *)

|   analyzeExp(let_in(_), typeEnv) = raise UnhandledLanguageConstruct
|   analyzeExp(anon_fun(_,_), typeEnv) = raise UnhandledLanguageConstruct

(*  
 * Not handled
 *)
|   analyzeExp(skip, typeEnv) = raise UnhandledLanguageConstruct
|   analyzeExp(case_of(_,_), typeEnv) = raise UnhandledLanguageConstruct
|   analyzeExp(exp_orelse(_,_), typeEnv) = raise UnhandledLanguageConstruct
|   analyzeExp(exp_andalso(_,_), typeEnv) = raise UnhandledLanguageConstruct
|   analyzeExp(curried_call(_,_), typeEnv) = raise UnhandledLanguageConstruct


(*****************************************************************************
 * Analyze a function definition.
 *****************************************************************************)


(*  
 * analyzeFunDef(fun_def, typeEnv): (typeExpr, substitutionMap)
 * 
 * Returns the inferred function type and an updated type environment for use
 * in analyzing later functions.
 *)
       
fun analyzeFunDef(fun_def(funName, funParams, funExp), typeEnv): 
                                                          TypeExp * TypeEnv =
       raise NotImplemented

(* print out the result of type inferencing a function *)

fun printFunType(fun_def(funName, _, _), typeExp) = 
       print("\nval " ^ funName ^ " = fn : " ^ typeToString(typeExp) ^ 
             "\n\n")


(*****************************************************************************
 * Analyze a whole program.
 *****************************************************************************)

(*
 * Analyze a list of declarations. The only declatations we alow for now are
 * function declataions. After each function declaration, print
 * out the type of the function. The return value is the type environment
 * after all the declarations.
 *)

fun analyzeListOfDec(nil, typeEnv) = typeEnv
|   analyzeListOfDec(non_structure_dec(fun_dec(fbind(fdef::nil)))::xs, 
                     typeEnv) =
       let
          val (typeExp, newTypeEnv) = analyzeFunDef(fdef, typeEnv);
       in
          (
             printFunType(fdef, typeExp);
             analyzeListOfDec(xs, newTypeEnv)
          )
       end
|   analyzeListOfDec(_, _) = raise UnhandledLanguageConstruct


(*
 * analyze a whole program in an environment
 *)

fun analyzeProg(prog(listOfDec), typeEnv) =
       analyzeListOfDec(listOfDec, typeEnv);


(*****************************************************************************
 * The user's main entry point to analyze a file
 *****************************************************************************)

fun test_type_infer(filename) =
       analyzeProg(MLParse(filename), createTopLevelEnv())


